/*
 * GCLC Socket, Socket implementation of GCLC
 * Copyright (C) 2014-2017 E. Bigeon
 * mailto:emmanuel@bigeon.fr
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
package fr.bigeon.gclc.socket;

import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.logging.Level;
import java.util.logging.Logger;

import fr.bigeon.gclc.ConsoleApplication;

/** This is a socket communicating console consoleManager.
 * <p>
 * To use this application, the following flow should be used:
 *
 * <pre>
 * SocketConsoleApplicationShell shell = new SocketConsoleApplicationShell();
 * ConsoleApplication myApplication = new MyConsoleApplication(shell.getConsoleManager(), ...);
 * shell.setApplication(myApplication);
 * Thread th = new Thread(shell);
 * th.start();
 * </pre>
 * <p>
 * This will start the application in a separate thread. The application will be
 * listening on the given socket and writing back on it. If this is all your
 * application, you should then {@link Thread#join()} the thread to wait for the
 * end of the execution.
 *
 * @author Emmanuel Bigeon */
public final class SocketConsoleApplicationShell implements Runnable {

    /** The class logger. */
    private static final Logger LOGGER = Logger
            .getLogger(SocketConsoleApplicationShell.class.getName());
    /** The listening port. */
    private final int port;
    /** The running status. */
    private boolean running;
    /** The socket console interface. */
    private SocketConsoleInterface sci;
    /** The remote disconnection command. */
    private ConnexionManager<Socket> rdc;
    /** The application. */
    private ConsoleApplication app;

    /** The server socket. */
    private ServerSocket serverSocket;
    /** THe server address. */
    private final InetAddress addr;

    /** Create a socket application shell which will listen on the given port
     * and network interface.
     *
     * @param port the part
     * @param addr the inet address */
    public SocketConsoleApplicationShell(final int port,
            final InetAddress addr) {
        super();
        this.port = port;
        this.addr = addr;
    }

    /** Wait for the identified connection to disconnect.
     *
     * @param id the connexion id. */
    private void awaitDisconnexion(final String id) {
        while (rdc.isConnected(id)) {
            try {
                rdc.waitDisconnexion(id);
            } catch (final InterruptedException e) {
                LOGGER.log(Level.SEVERE, "Unexpected interruption", e); //$NON-NLS-1$
                Thread.currentThread().interrupt();
            }
        }
        sci.disconnect();
    }

    /** If the port provided was 0, this allows to get the actual port.
     * @return the local port
     * @see java.net.ServerSocket#getLocalPort()
     */
    public int getLocalPort() {
        return serverSocket.getLocalPort();
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run() */
    @Override
    public void run() {
        // Create the server
        try (ServerSocket actualServerSocket = new ServerSocket(port, 1,
                addr)) {
            serverSocket = actualServerSocket;
            running = true;
            // Create the streams
            runSokectServer();
        } catch (final IOException e) {
            LOGGER.severe("Communication error between client and server"); //$NON-NLS-1$
            LOGGER.log(Level.FINE,
                    "Communication error between client and server", e); //$NON-NLS-1$
        }
    }

    /** Acctually run the server loop on connexion.
     *
     * @throws IOException if the communication with the client failed */
    private void runSokectServer() throws IOException {
        while (running && app.isRunning()) {
            LOGGER.info("Waiting client"); //$NON-NLS-1$
            try (Socket clientSocket = serverSocket.accept();) {

                sci.connect(clientSocket);

                final String id = rdc.addConnexion(clientSocket);
                rdc.lockDisconnexion(id);
                awaitDisconnexion(id);
                rdc.releaseDisconnexionLock(id);
            } catch (final SocketException e) {
                LOGGER.log(Level.INFO, "Socket closed"); //$NON-NLS-1$
                LOGGER.log(Level.FINE,
                        "Socket closed with exception (probably due to server interruption)", //$NON-NLS-1$
                        e);
            } catch (final IOException e) {
                throw e;
            }
            LOGGER.info("Closing client"); //$NON-NLS-1$
        }
        LOGGER.info("Closing Server"); //$NON-NLS-1$
    }

    /** Set the application.
     * <p>
     * If the application is closed, the server will also close.
     *
     * @param app the app to set */
    public void setApplication(final ConsoleApplication app) {
        this.app = app;
    }

    /** Set the connexion manager.
     *
     * @param rdc the rdc to set */
    public void setConnexionManager(final ConnexionManager<Socket> rdc) {
        this.rdc = rdc;
    }

    /** Set the socket console interface.
     *
     * @param sci the console interface to set */
    public void setInterface(final SocketConsoleInterface sci) {
        this.sci = sci;
    }

    /** This method will request the server to stop.
     * <p>
     * In most cases, this will terminate communication on every client. On some
     * cases, the */
    public void stop() {
        running = false;
        try {
            serverSocket.close();
        } catch (final IOException e) {
            LOGGER.log(Level.SEVERE, "Exception in closing socket server", e); //$NON-NLS-1$
        }
    }
}
