/*
 * GCLC swt, provide a swt window for console applications
 * Copyright (C) 2015-2017 E. Bigeon
 * mailto:emmanuel@bigeon.fr
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc-swt:fr.bigeon.gclc.swt.SWTConsole.java
 * Created on: Apr 18, 2015
 */
package fr.bigeon.gclc.swt;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;

import fr.bigeon.gclc.ConsoleApplication;
import fr.bigeon.gclc.manager.ConsoleInput;
import fr.bigeon.gclc.manager.ConsoleOutput;
import fr.bigeon.gclc.tools.ConstantString;
import fr.bigeon.gclc.tools.StringProvider;

/** A SWT component to connect to gclc {@link ConsoleApplication}.
 * <p>
 *
 * @author Emmanuel Bigeon */
public final class SWTConsole extends Composite
        implements ConsoleDelayIO, ConsoleInput, ConsoleOutput {
    /** The number of columns of the layout. */
    private static final int    LAYOUT_NB_COLUMNS = 2;
    /** The cmd prefix in the output console. */
    private static final String CMD_PREFIX        = "[CMD] ";                 //$NON-NLS-1$
    /** The class logger. */
    private static final Logger LOGGER            = Logger
            .getLogger(SWTConsole.class.getName());
    /** The empty string constant. */
    private static final String EMPTY             = "";                       //$NON-NLS-1$
    /** The console output text field. */
    private final Text          consoleOutput;
    /** The console input text field. */
    private final Text          consoleInput;
    /** The prompt label. */
    private final Label         lblPromptlabel;
    /** The prompt text. */
    private StringProvider      prompt            = new ConstantString("> "); //$NON-NLS-1$
    /** The command entered by the user. */
    private String              command           = null;
    /** If the prompt should be active. */
    private boolean             prompting         = false;
    /** The object for thread synchronization with the prompt. */
    private final Object        promptLock        = new Object();

    /** Create the composite.
     *
     * @param parent the prent composite
     * @param style the composite style */
    public SWTConsole(final Composite parent, final int style) {
        super(parent, style);

        setLayout(new GridLayout(LAYOUT_NB_COLUMNS, false));

        consoleOutput = new Text(this,
                SWT.BORDER | SWT.READ_ONLY | SWT.WRAP | SWT.V_SCROLL | SWT.MULTI);
        consoleOutput.setLayoutData(
                new GridData(SWT.FILL, SWT.FILL, true, true, LAYOUT_NB_COLUMNS, 1));
        consoleOutput.setRedraw(true);

        lblPromptlabel = new Label(this, SWT.NONE);
        lblPromptlabel.setText(prompt.apply());

        consoleInput = new Text(this, SWT.BORDER);
        consoleInput.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
        consoleInput.addKeyListener(new HistoryTextKeyListener(this));

    }

    @Override
    protected void checkSubclass() {
        // Disable the check that prevents subclassing of SWT components
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleManager#close() */
    @Override
    public void close() {
        synchronized (promptLock) {
            promptLock.notifyAll();
        }
        if (consoleInput.isDisposed()) {
            return;
        }
        consoleInput.setEnabled(false);
        consoleOutput.setEnabled(false);
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.swt.ConsoleDelayIO#getInput() */
    @Override
    public String getInput() {
        return consoleInput.getText();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#getPrompt() */
    @Override
    public StringProvider getPrompt() {
        return prompt;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleManager#interruptPrompt() */
    @Override
    public void interruptPrompt() {
        synchronized (promptLock) {
            promptLock.notifyAll();
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleManager#isClosed() */
    @Override
    public boolean isClosed() {
        return isDisposed();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#print(java.lang.String) */
    @Override
    public void print(final String text) {
        Display.getDefault().syncExec(new Runnable() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                if (!consoleOutput.isDisposed()) {
                    consoleOutput.append(text);
                }
            }
        });
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#println() */
    @Override
    public void println() {
        Display.getDefault().syncExec(new Runnable() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                if (!consoleOutput.isDisposed()) {
                    consoleOutput.append(System.lineSeparator());
                }
            }
        });
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#println(java.lang.String) */
    @Override
    public void println(final String message) {
        Display.getDefault().syncExec(new Runnable() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                if (!consoleOutput.isDisposed()) {
                    consoleOutput.append(message + System.lineSeparator());
                }
            }
        });

    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#prompt() */
    @Override
    public String prompt() throws IOException {
        synchronized (promptLock) {
            if (isDisposed()) {
                throw new IOException();
            }
            try {
                Display.getDefault().syncExec(new Runnable() {
                    @Override
                    public void run() {
                        if (!consoleInput.isDisposed()) {
                            consoleInput.setEnabled(true);
                            lblPromptlabel.setText(prompt.apply());
                            // relayout
                            SWTConsole.this.layout();
                            consoleInput.setFocus();
                        }
                    }
                });
                prompting = true;
                promptLock.notifyAll();
                while (prompting) {
                    promptLock.wait();
                }
            } catch (final InterruptedException e) {
                LOGGER.log(Level.WARNING, "Error in synchronization of prompting", e); //$NON-NLS-1$
                command = null;
                Thread.currentThread().interrupt();
            }
        }
        if (isDisposed()) {
            throw new IOException("Input closed"); //$NON-NLS-1$
        }
        return command;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(long) */
    @Override
    public String prompt(final long timeout) throws IOException {
        synchronized (promptLock) {
            if (isDisposed()) {
                throw new IOException();
            }
            try {
                Display.getDefault().syncExec(new Runnable() {
                    @Override
                    public void run() {
                        if (!consoleInput.isDisposed()) {
                            consoleInput.setEnabled(true);
                            lblPromptlabel.setText(prompt.apply());
                            // relayout
                            SWTConsole.this.layout();
                            consoleInput.setFocus();
                        }
                    }
                });
                prompting = true;
                command = null;
                promptLock.notifyAll();
                long start = System.currentTimeMillis();
                long cur = start;
                while (prompting && start + timeout>cur) {
                    promptLock.wait((cur-start-timeout)/2);
                    cur = System.currentTimeMillis();
                }
            } catch (final InterruptedException e) {
                LOGGER.log(Level.WARNING, "Error in synchronization of prompting", e); //$NON-NLS-1$
                command = null;
                Thread.currentThread().interrupt();
            }
        }
        if (isDisposed()) {
            throw new IOException("Input closed"); //$NON-NLS-1$
        }
        return command;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#prompt(java.lang.String) */
    @Override
    public String prompt(final String message) throws IOException {
        synchronized (promptLock) {
            if (isDisposed()) {
                throw new IOException();
            }
            try {
                Display.getDefault().syncExec(new Runnable() {
                    @SuppressWarnings("synthetic-access")
                    @Override
                    public void run() {
                        if (!consoleOutput.isDisposed()) {
                            lblPromptlabel.setText(message);
                            // relayout
                            SWTConsole.this.layout();
                            consoleInput.setEnabled(true);
                            consoleInput.setFocus();
                        }
                    }
                });
                prompting = true;
                promptLock.wait();
                if (isDisposed()) {
                    throw new IOException();
                }
            } catch (final InterruptedException e) {
                LOGGER.log(Level.WARNING, "Error in synchronization of prompting", e); //$NON-NLS-1$
                command = null;
                Thread.currentThread().interrupt();
            } finally {
                Display.getDefault().syncExec(new Runnable() {
                    @SuppressWarnings("synthetic-access")
                    @Override
                    public void run() {
                        if (!consoleOutput.isDisposed()) {
                            lblPromptlabel.setText(prompt.apply());
                            // relayout
                            SWTConsole.this.layout();
                        }
                    }
                });
            }
        }
        return command;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.manager.ConsoleInput#prompt(java.lang.String, long) */
    @Override
    public String prompt(final String message, final long timeout) throws IOException {
        // TODO Auto-generated method stub
        // return null;
        throw new RuntimeException("Not implemented yet");
    }

    /* (non-Javadoc)
     * @see org.eclipse.swt.widgets.Composite#setFocus() */
    @Override
    public boolean setFocus() {
        return consoleInput.setFocus();
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.swt.ConsoleDelayIO#setInput(java.lang.String) */
    @Override
    public void setInput(final String input) {
        consoleInput.setText(input);
        consoleInput.setSelection(input.length());
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.ConsoleManager#setPrompt(java.lang.String) */
    @Override
    public void setPrompt(final StringProvider prompt) {
        this.prompt = prompt;
        Display.getDefault().syncExec(new Runnable() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                if (!consoleOutput.isDisposed()) {
                    lblPromptlabel.setText(prompt.apply());
                    // relayout
                    SWTConsole.this.layout();
                }
            }
        });
    }

    @Override
    public void setPrompt(String prompt) {
        setPrompt(new ConstantString(prompt));
    }

    /** @param string the text */
    public void setText(final String string) {
        consoleInput.setText(string);
    }

    /**
     *
     */
    public void validateCommand() {
        validateInput();
    }

    /**
     *
     */
    @Override
    public void validateInput() {
        Display.getDefault().syncExec(new Runnable() {
            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                consoleInput.setEnabled(false);
            }
        });
        synchronized (promptLock) {
            while (!prompting) {
                try {
                    promptLock.wait();
                } catch (final InterruptedException e) {
                    LOGGER.log(Level.SEVERE, "Interruption while waiting prompt", e); //$NON-NLS-1$
                }
            }
            Display.getDefault().syncExec(new Runnable() {
                @SuppressWarnings("synthetic-access")
                @Override
                public void run() {
                    command = consoleInput.getText();
                    prompting = false;
                    consoleInput.setText(EMPTY);
                    consoleOutput.append(CMD_PREFIX + command + System.lineSeparator());
                }
            });
            promptLock.notifyAll();
        }
    }

}
