/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/** gclc:fr.bigeon.gclc.command.SubedCommand.java
 * Created on: Sep 6, 2014 */
package fr.bigeon.gclc.command;

import java.io.IOException;
import java.util.Arrays;

import fr.bigeon.gclc.exception.CommandRunException;
import fr.bigeon.gclc.exception.CommandRunExceptionType;
import fr.bigeon.gclc.manager.ConsoleInput;
import fr.bigeon.gclc.manager.ConsoleOutput;

/**
 * <p>
 * A subed command is a command that can execute sub commands depending on the
 * first argument.
 *
 * @author Emmanuel BIGEON */
public final class SubedCommand extends CommandProvider implements ICommand {

    /** The tab character. */
    private static final String TAB = "\t"; //$NON-NLS-1$
    /** The command to execute when this command is called with no sub
     * arguments.
     * <p>
     * This may be null, in which case the command should have arguments. */
    private final ICommand noArgCommand;
    /** A tip on this command. */
    private final String tip;
    /** The name of the command. */
    private final String name;

    /** Create the command that defines sub commands.
     *
     * @param name the name of the command */
    public SubedCommand(final String name) {
        super();
        this.name = name;
        noArgCommand = null;
        tip = null;
    }

    /** Create the command that defines sub commands.
     *
     * @param name the name of the command
     * @param noArgCommand the command to execute when no extra parameter are
     *            provided */
    public SubedCommand(final String name, final ICommand noArgCommand) {
        super();
        this.name = name;
        this.noArgCommand = noArgCommand;
        tip = null;
    }

    /** Create the command that defines sub commands.
     *
     * @param name the name of the command
     * @param noArgCommand the command to execute
     * @param tip the help tip associated */
    public SubedCommand(final String name, final ICommand noArgCommand,
            final String tip) {
        super();
        this.name = name;
        this.noArgCommand = noArgCommand;
        this.tip = tip;
    }

    /** Create the command that defines sub commands.
     *
     * @param name the name of the command
     * @param tip the help tip associated */
    public SubedCommand(final String name, final String tip) {
        super();
        this.name = name;
        noArgCommand = null;
        this.tip = tip;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.acide.Command#execute(java.lang.String[]) */
    @Override
    public void execute(final ConsoleOutput output, final ConsoleInput input,
                        final String... args) throws CommandRunException {
        if (args.length == 0 || args[0].startsWith("-")) { //$NON-NLS-1$
            if (noArgCommand != null) {
                noArgCommand.execute(output, input, args);
            } else {
                throw new CommandRunException("Unrecognized command", this); //$NON-NLS-1$
            }
        } else {

            try {
                executeSub(output, input, args[0],
                        Arrays.copyOfRange(args, 1, args.length));
            } catch (final CommandRunException e) {
                if (e.getSource() != null) {
                    throw e;
                }
                throw new CommandRunException(CommandRunExceptionType.USAGE,
                        e.getLocalizedMessage(), e, this);
            }
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.ICommand#getCommandName() */
    @Override
    public String getCommandName() {
        return name;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.Command#help() */
    @Override
    public void help(final ConsoleOutput manager,
                     final String... args) throws IOException {
        if (args.length != 0 && !args[0].startsWith("-")) { //$NON-NLS-1$
            // Specific
            final ICommand c = get(args[0]);
            if (c != null) {
                c.help(manager, Arrays.copyOfRange(args, 1, args.length));
            } else {
              manager.println("No command "+Arrays.toString(args));
            }
        } else {
            // Generic
            if (noArgCommand != null && noArgCommand.tip() != null) {
                manager.println(TAB + noArgCommand.tip());
            }
            for (final ICommand cmd : commands) {
                if (cmd.tip() == null) {
                    manager.println(TAB + cmd.getCommandName());
                } else {
                    manager.println(TAB + cmd.getCommandName() + ": " + //$NON-NLS-1$
                                    cmd.tip());
                }
            }
        }
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.Command#tip() */
    @Override
    public String tip() {
        return tip;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#toString() */
    @Override
    public String toString() {
        return "SubedCommand " + super.toString(); //$NON-NLS-1$
    }
}
